"""
========================================================================
Component.py
========================================================================
Add clk/reset signals.

Author : Yanghui Ou
  Date : Apr 6, 2019
"""

from .ComponentLevel1 import ComponentLevel1
from .ComponentLevel7 import ComponentLevel7
from .Connectable import (
    Connectable,
    Const,
    InPort,
    Interface,
    MethodPort,
    OutPort,
    Signal,
    Wire,
)
from .errors import (
    InvalidAPICallError,
    InvalidConnectionError,
    NotElaboratedError,
    UnsetMetadataError,
)
from .MetadataKey import MetadataKey
from .NamedObject import NamedObject
from .Placeholder import Placeholder


class Component( ComponentLevel7 ):

  #-----------------------------------------------------------------------
  # Private methods
  #-----------------------------------------------------------------------

  def __new__( cls, *args, **kwargs ):
    inst = super().__new__( cls, *args, **kwargs )
    # Maps a MetadataKey instance to its value
    inst._metadata = {}
    return inst

  # Override
  def _construct( s ):

    if not s._dsl.constructed:

      # clk and reset signals are added here.
      s.clk   = InPort()
      s.reset = InPort()

      # Merge the actual keyword args and those args set by set_parameter
      if s._dsl.param_tree is None:
        kwargs = s._dsl.kwargs
      elif s._dsl.param_tree.leaf is None:
        kwargs = s._dsl.kwargs
      else:
        kwargs = s._dsl.kwargs
        if "construct" in s._dsl.param_tree.leaf:
          more_args = s._dsl.param_tree.leaf[ "construct" ]
          kwargs.update( more_args )

      s._handle_decorated_methods()

      # Same as parent class _construct
      s.construct( *s._dsl.args, **kwargs )

      # We hook up the added clk and reset signals here. NOTE THAT if the
      # user overwrites clk/reset inside the component, we still get the
      # correct connection.
      parent = s.get_parent_object()
      if parent is not None:
        parent._connect_signal_signal( s.clk, parent.clk )
        parent._connect_signal_signal( s.reset, parent.reset )

      s._dsl.constructed = True

  # This function deduplicates those checks in each API
  def _check_called_at_elaborate_top( s, func_name ):
    try:
      if s._dsl.elaborate_top is not s:
        raise InvalidAPICallError( func_name, s, s._dsl.elaborate_top )
    except AttributeError:
      raise NotElaboratedError()

  def _collect_objects_local( s, filt, sort_key = None ):
    assert s._dsl.constructed
    ret = set()
    stack = []
    for (name, obj) in s.__dict__.items():
      if   isinstance( name, str ):
        if name[0] != '_': # filter private variables
          stack.append( obj )
    while stack:
      u = stack.pop()
      if filt( u ):
        ret.add( u )
      # ONLY LIST IS SUPPORTED
      elif isinstance( u, list ):
        stack.extend( u )
    if sort_key:
      return sorted( ret, key = sort_key )
    else:
      return list(ret)

  def _flush_pending_value_connections( s ):
    if s._dsl._has_pending_value_connections:
      s._dsl.all_value_nets = s._resolve_value_connections()
      s._dsl._has_pending_value_connections = False

  def _flush_pending_method_connections( s ):
    if s._dsl._has_pending_method_connections:
      s._dsl.all_method_nets = s._resolve_method_connections()
      s._dsl._has_pending_method_connections = False

  # These internal functions are implemented in a more generic way. The
  # public APIs should wrap around these functions.

  def _add_component( s, parent, name, indices, obj, provided_connections,
                      provided_upblk_reads, provided_upblk_writes, provided_upblk_calls,
                      provided_func_reads,  provided_func_writes, provided_func_calls ):
    try:
      top = s._dsl.elaborate_top
    except AttributeError:
      raise NotElaboratedError()

    NamedObject._elaborate_stack = [ parent ]

    # Check if we are adding obj to a list of to a component
    if not indices:
      # If we are adding field parent.x, we simply reuse the setattr hook
      assert not hasattr( parent, name ), f"Invalid add_component call: {parent} already has field {name}!"

      NamedObject.__setattr__ = NamedObject.__setattr_for_elaborate__
      setattr( parent, name, obj )
      del NamedObject.__setattr__

    else:
      assert hasattr( parent, name ), f"Invalid add_component call: Field '{name}' of {parent} should exist and be a list, but doesn't exist!"

      # We use the indices to get the list parent and set the element

      list_parent = getattr( parent, name )
      i = 0
      while i < len(indices) - 1:
        list_parent = list_parent[ indices[i] ]
        i += 1

      assert list_parent[ indices[i] ] is None, f"Invalid add_component call: {parent} already has field '{name}!'"
      list_parent[ indices[i] ] = obj

      # NOTE THAT we cannot use setattr hook for the top level metadata
      # because setting an element in the list doesn't call setattr.
      # As a result we have to copy some code from setattr hook and do it
      # manually here for the top level

      obj._dsl.parent_obj = parent
      obj._dsl.level      = parent._dsl.level + 1
      obj._dsl.my_name    = u_name = name + "".join( [ f"[{x}]" for x in indices ] )

      # Iterate through the param_tree and update u
      if parent._dsl.param_tree is not None:
        if parent._dsl.param_tree.children is not None:
          for comp_name, node in parent._dsl.param_tree.children.items():
            if comp_name == u_name:
              # Lazily create the param tree
              if obj._dsl.param_tree is None:
                obj._dsl.param_tree = ParamTreeNode()
              obj._dsl.param_tree.merge( node )

            elif node.compiled_re is not None:
              if node.compiled_re.match( u_name ):
                # Lazily create the param tree
                if obj._dsl.param_tree is None:
                  obj._dsl.param_tree = ParamTreeNode()
                obj._dsl.param_tree.merge( node )

      obj._dsl.full_name = ( parent._dsl.full_name + "." + u_name )

      # store the name/indices
      obj._dsl._my_name     = name
      obj._dsl._my_indices  = indices

      obj._dsl.elaborate_top = top
      obj._dsl.NamedObject_fields = set()

      NamedObject._elaborate_stack.append( obj )
      NamedObject.__setattr__ = NamedObject.__setattr_for_elaborate__
      obj._construct()
      del NamedObject.__setattr__

      NamedObject._elaborate_stack.pop()

    added_components = obj._collect_all_single( lambda x: isinstance( x, Component ) )

    # First elaborate all functions to spawn more named objects
    for c in added_components:
      c._elaborate_read_write_func()

    added_signals, added_method_ports = \
      obj._collect_all( [ lambda x: isinstance( x, Signal ), \
                          lambda x: isinstance( x, MethodPort ) ] )

    top._dsl.all_components    |= added_components
    top._dsl.all_signals       |= added_signals
    top._dsl.all_method_ports  |= added_method_ports

    top._dsl.all_named_objects |= added_components
    top._dsl.all_named_objects |= added_signals
    top._dsl.all_named_objects |= added_method_ports

    for c in added_components:
      top._collect_vars( c )

    # Lazy -- to avoid resolve_connection call which takes non-trivial
    # time upon adding any connect, I just mark pending here. Whenever you
    # call the right API which is get_all_value_nets()/get_method_nets(),
    # the API will batch flush all pending connections if the flag is
    # true. So please make sure to call the API whenever you want to get
    # the value/method nets.
    connection_pairs = []
    for (x, y) in provided_connections:
      connection_pairs.append( x )
      connection_pairs.append( eval(y) )
      if not top._dsl._has_pending_value_connections and isinstance( x, Signal ):
        top._dsl._has_pending_value_connections = True
      if not top._dsl._has_pending_method_connections and isinstance( x, MethodPort ):
        top._dsl._has_pending_method_connections = True

    # WE NEED TO ADD CONNECTIONS AT PARENT INSTEAD OF TOP
    parent.add_connections( *connection_pairs )

    # Now we put back the provided upblk metadata to parent and top
    for blk, obj_name in provided_upblk_reads:
      parent._dsl.upblk_reads[blk].add( eval(obj_name) )

    for blk, obj_name in provided_upblk_writes:
      parent._dsl.upblk_writes[blk].add( eval(obj_name) )

    for blk, obj_name in provided_upblk_calls:
      parent._dsl.upblk_calls[blk].add( eval(obj_name) )

    for func, obj_name in provided_func_reads:
      parent._dsl.func_reads[func].add( eval(obj_name) )

    for func, obj_name in provided_func_writes:
      parent._dsl.func_writes[func].add( eval(obj_name) )

    for func, obj_name in provided_func_calls:
      parent._dsl.func_calls[func].add( eval(obj_name) )

    del NamedObject._elaborate_stack

  def _delete_component( top, obj ):

    # This nested delete function is to create an extra layer to properly
    # call garbage collector. If we can make sure it is collected
    # automatically and fast enough, we might remove this extra layer
    #
    # EDIT: After experimented w/ and w/o gc.collect(), it seems like it
    # is eventually collected, but sometimes the intermediate memory
    # footprint might reach up to gigabytes, so let's keep the
    # gc.collect() for now

    def _delete_component_internal( top, foo ):
      assert obj._dsl.elaborate_top is top

      # First make sure we flush pending connections
      value_nets  = top.get_all_value_nets()
      method_nets = top.get_all_method_nets()

      # Delete the placeholder from parent. We need to see if we are
      # dealing with a list parent or a component parent.

      parent = foo.get_parent_object()

      if foo._dsl._my_indices:
        # We use the saved indices to get the list parent and set the index
        my_indices = foo._dsl._my_indices
        list_parent = getattr( parent, foo._dsl._my_name )
        i = 0
        while i < len(my_indices) - 1:
          list_parent = list_parent[ my_indices[i] ]
          i += 1
        list_parent[ my_indices[i] ] = None
      else:
        # Simply delete the attribute if it's merely a field
        delattr( parent, foo._dsl.my_name )
        parent._dsl.NamedObject_fields.remove( foo._dsl.my_name )

      # Remove all components, signals, and method ports
      removed_components, removed_signals, removed_method_ports = \
        foo._collect_all( [ lambda x: isinstance( x, Component ), \
                            lambda x: isinstance( x, Signal ), \
                            lambda x: isinstance( x, MethodPort ) ] )

      top._dsl.all_components    -= removed_components
      top._dsl.all_signals       -= removed_signals
      top._dsl.all_method_ports  -= removed_method_ports

      top._dsl.all_named_objects -= removed_components

      removed_connectables = removed_signals | removed_method_ports
      top._dsl.all_named_objects -= removed_connectables

      removed_consts = set()
      if isinstance( foo, Placeholder ):
        # No need to uncollect vars from a placeholder
        assert not foo._dsl.consts
      else:
        for x in removed_components:
          # remove consts
          removed_consts |= x._dsl.consts
          # uncollect variables
          top._uncollect_vars( x )

      saved_upblk_reads  = []
      saved_upblk_writes = []
      saved_upblk_calls  = []
      saved_func_reads   = []
      saved_func_writes  = []
      saved_func_calls   = []

      # If an update block/function in parent writes the inport or reads
      # the outport or calls a callee method of the deleted component, we
      # must save the information (upA reads B) to avoid bugs or
      # explicitly re-elaborating the parent.

      for blk, reads in parent._dsl.upblk_reads.items():
        assert blk in top._dsl.all_upblk_reads
        to_save = set()
        for x in reads:
          if x in removed_connectables:
            to_save.add( x )
            saved_upblk_reads.append( (blk, repr(x)) )
        parent._dsl.upblk_reads[blk] -= to_save

      for blk, writes in parent._dsl.upblk_writes.items():
        assert blk in top._dsl.all_upblk_writes
        to_save = set()
        for x in writes:
          if x in removed_connectables:
            to_save.add( x )
            saved_upblk_writes.append( (blk, repr(x)) )
        parent._dsl.upblk_writes[blk] -= to_save

      for blk, calls in parent._dsl.upblk_calls.items():
        assert blk in top._dsl.all_upblk_calls
        to_save = set()
        for x in calls:
          if x in removed_connectables:
            to_save.add( x )
            saved_upblk_calls.append( (blk, repr(x)) )
        parent._dsl.upblk_calls[blk] -= to_save

      # We need to save the information for funcs too
      for func, reads in parent._dsl.func_reads.items():
        to_save = set()
        for x in reads:
          if x in removed_connectables:
            to_save.add( x )
            saved_func_reads.append( (func, repr(x)) )
        parent._dsl.func_reads[func] -= to_save

      for func, writes in parent._dsl.func_writes.items():
        to_save = set()
        for x in writes:
          if x in removed_connectables:
            to_save.add( x )
            saved_func_writes.append( (func, repr(x)) )
        parent._dsl.func_writes[func] -= to_save

      for func, calls in parent._dsl.func_calls.items():
        to_save = set()
        for x in calls:
          if x in removed_connectables:
            to_save.add( x )
            saved_func_calls.append( (func, repr(x)) )
        parent._dsl.func_calls[func] -= to_save

      saved_connections = []

      for x in removed_connectables:
        # Clean up all_adjancency at top
        if x in top._dsl.all_adjacency:
          for other in top._dsl.all_adjacency[x]:
            # other must be in the dict
            # If other will be removed, we don't need to remove it here ..
            if other not in removed_connectables and other not in removed_consts:
              top._dsl.all_adjacency[other].remove( x )
              if isinstance( other, Const ):
                other = other._dsl.const
              saved_connections.append( (other, "top"+repr(x)[1:]) ) # other is from outside
          del top._dsl.all_adjacency[x]

        # Clean up adjacency at parent
        if x in parent._dsl.adjacency:
          for other in parent._dsl.adjacency[x]:
            # other must be in the dict
            if other not in removed_connectables:
              parent._dsl.adjacency[other].remove( x )
          del parent._dsl.adjacency[x]

      for x in removed_components:
        del x._dsl.parent_obj
        del x._dsl.elaborate_top
      for x in removed_connectables:
        del x._dsl.parent_obj
        del x._dsl.elaborate_top
        x._dsl.full_name = "<deleted>"+x._dsl.full_name
      for y in removed_consts:
        del y._dsl.parent_obj

      # We don't break nets anymore. Instead, we set the flags to true so
      # that the next get_xxx_net will immediately recollect nets.
      top._dsl._has_pending_value_connections = True
      top._dsl._has_pending_method_connections = True

      # We clean up the connect_order list. If we want to preserve the
      # original connect order, we can play some other tricks here such as
      # leaving the strings in there. However I don't think that's clean
      # enough. Thus I'm just removing them right now.
      new_connect_order = []
      for (x, y) in parent._dsl.connect_order:
        if x not in removed_signals and y not in removed_signals: # TODO method port
          new_connect_order.append( (x, y) )

      parent._dsl.connect_order = new_connect_order

      return saved_connections, saved_upblk_reads, saved_upblk_writes, saved_upblk_calls, \
             saved_func_reads, saved_func_writes, saved_func_calls

    return _delete_component_internal( top, obj )
    # import gc
    # gc.collect() # this takes 0.1 seconds

  # Override, add pypy hooks
  def elaborate( s ):
    try:
      import pypyjit
      pypyjit.set_param("off")
    except:
      pass

    super().elaborate()

    # try:
      # import pypyjit
      # pypyjit.set_param("default")
    # except:
      # pass

  #-----------------------------------------------------------------------
  # Public APIs (can be called either before or after elaboration)
  #-----------------------------------------------------------------------

  """ Metadata APIs """

  def set_metadata( s, key, value ):
    """Set the metadata ``key`` of the given component to be ``value``.

    Can be called before, during, or after elaboration.

    Args:
        key (MetadataKey): Key of the metadata.
        value (object): The metadata. Can be any object.
    """
    key.check_value( value )
    s._metadata[ key ] = value

  def has_metadata( s, key ):
    """Check if the component has metadata for ``key``.

    Can be called before, during, or after elaboration.

    Args:
        key (MetadataKey): Key of the metadata.

    Returns:
        bool: Whether or not the component has the metadata for ``key``.

    Raises:
        TypeError: Raised if ``key`` is not an instance of :class:`MetadataKey`.
    """
    if not isinstance( key, MetadataKey ):
      raise TypeError(f'the given object {key} is not a MetadataKey!')
    return key in s._metadata

  def get_metadata( s, key ):
    """Get the metadata ``key`` of the given component.

    Can be called before, during, or after elaboration.

    Args:
        key (MetadataKey): Key of the metadata.

    Returns:
        object: The metadata of the given ``key``.

    Raises:
        TypeError: Raised if ``key`` is not an instance of :class:`MetadataKey`.
        UnsetMetadataError: Raised if the component does not have metadata for
            the given ``key``.
    """
    if not s.has_metadata( key ):
      raise UnsetMetadataError( key, s )
    return s._metadata[ key ]

  #-----------------------------------------------------------------------
  # Post-elaborate public APIs (can only be called after elaboration)
  #-----------------------------------------------------------------------

  """ Convenience/utility APIs """

  def apply( s, pass_instance ):
    try:
      import pypyjit
      pypyjit.set_param("off")
    except:
      pass
    if not hasattr( s._dsl, "elaborate_top" ):
      s.elaborate()

    assert type(pass_instance) is not type, f"Should pass in a pass instance like " \
                                            f"'{pass_instance.__name__}()' instead of '{pass_instance.__name__}'"
    assert callable( pass_instance ), f"Should override __call__ of {pass_instance.__name__} for a valid pass"
    pass_instance( s )

  def check( s ):
    s._check_valid_dsl_code()

  """ APIs that provide local metadata of a component """

  def get_component_level( s ):
    try:
      return s._dsl.level
    except AttributeError:
      raise NotElaboratedError()

  def get_child_components( s, sort_key = None ):
    return s._collect_objects_local( lambda x: isinstance( x, Component ), sort_key )

  def get_input_value_ports( s, sort_key = None ):
    return s._collect_objects_local( lambda x: isinstance( x, InPort ), sort_key )

  def get_output_value_ports( s , sort_key = None ):
    return s._collect_objects_local( lambda x: isinstance( x, OutPort ), sort_key )

  def get_wires( s , sort_key = None ):
    return s._collect_objects_local( lambda x: isinstance( x, Wire ), sort_key )

  def get_update_blocks( s ):
    assert s._dsl.constructed
    return s._dsl.upblks

  def get_update_ff( s ):
    assert s._dsl.constructed
    return s._dsl.update_ff

  def get_upblk_metadata( s ):
    assert s._dsl.constructed
    return s._dsl.upblk_reads, s._dsl.upblk_writes, s._dsl.upblk_calls

  # These xxx_order APIs should be used when some pass wants to process
  # connections/upblks in the order that they are created

  def get_connect_order( s ):
    try:
      return s._dsl.connect_order
    except AttributeError:
      raise NotElaboratedError()

  def get_update_block_order( s ):
    try:
      return s._dsl.upblk_order
    except AttributeError:
      raise NotElaboratedError()

  """ Metadata APIs that should only be called at elaborated top"""

  def get_all_object_filter( s, filt ):
    assert callable( filt )
    try:
      return { x for x in s._dsl.all_named_objects if filt(x) }
    except AttributeError:
      return s._collect_all_single( filt )

  def get_local_object_filter( s, filt, sort_key = None ):
    assert callable( filt )
    return s._collect_objects_local( filt, sort_key )

  def get_all_components( s ):
    try:
      s._check_called_at_elaborate_top( "get_all_update_blocks" )
      return s._dsl.all_components
    except AttributeError:
      raise NotElaboratedError()

  def get_all_update_blocks( s ):
    try:
      s._check_called_at_elaborate_top( "get_all_update_blocks" )
      return s._dsl.all_upblks
    except AttributeError:
      raise NotElaboratedError()

  def get_all_update_ff( s ):
    try:
      s._check_called_at_elaborate_top( "get_all_update_ff" )
      return s._dsl.all_update_ff
    except AttributeError:
      raise NotElaboratedError()

  def get_all_update_once( s ):
    try:
      s._check_called_at_elaborate_top( "get_all_update_once" )
      return s._dsl.all_update_once
    except AttributeError:
      raise NotElaboratedError()

  def get_all_upblk_metadata( s ):
    try:
      s._check_called_at_elaborate_top( "get_all_upblk_metadata" )
      return s._dsl.all_upblk_reads, s._dsl.all_upblk_writes, s._dsl.all_upblk_calls
    except AttributeError:
      raise NotElaboratedError()

  def get_all_explicit_constraints( s ):
    try:
      s._check_called_at_elaborate_top( "get_all_explicit_constraints" )
      return s._dsl.all_U_U_constraints, s._dsl.all_RD_U_constraints, \
             s._dsl.all_WR_U_constraints, s._dsl.all_M_constraints
    except AttributeError:
      raise NotElaboratedError()

  # is_lambda?, src, line, filename, ast
  def get_update_block_info( s, blk ):
    try:
      name_info = s.__class__._name_info
    except AttributeError: # This component doesn't have update block
      return None

    return name_info[blk.__name__]

  def get_update_block_host_component( s, blk ):
    try:
      s._check_called_at_elaborate_top( "get_update_block_host_component" )
      return s._dsl.all_upblk_hostobj[ blk ]
    except AttributeError:
      raise NotElaboratedError()

  def get_all_method_nets( s ):
    s._check_called_at_elaborate_top( "get_all_method_nets" )
    s._flush_pending_method_connections()
    return s._dsl.all_method_nets

  # Override
  def get_all_value_nets( s ):
    s._check_called_at_elaborate_top( "get_all_value_nets" )
    s._flush_pending_value_connections()
    return s._dsl.all_value_nets

  def get_signal_adjacency_dict( s ):
    try:
      s._check_called_at_elaborate_top( "get_signal_adjacency_dict" )
    except AttributeError:
      raise NotElaboratedError()
    return s._dsl.all_adjacency

  """ Mutation APIs to add/delete components and connections"""

  # Shunning: These API implement replacing a component with another
  # module. This is the first mutation API we want to test thoroughly

  # TODO test everything below and figure out whether we need to delete
  # a normal component

  def replace_component( top, foo, cls, check=True ):
    top._check_called_at_elaborate_top( "replace_component" )

    parent = foo.get_parent_object()
    foo_name    = foo._dsl._my_name
    foo_indices = foo._dsl._my_indices

    saved_connections, saved_upblk_reads, saved_upblk_writes, saved_upblk_calls, \
      saved_func_reads, saved_func_writes, saved_func_calls = top._delete_component( foo )

    new_obj = cls( *foo._dsl.args, **foo._dsl.kwargs )

    # We actually don't need to merge param tree here because when we call
    # _add_component, the parameters stored in parent will be pushed down
    # to new_obj
    top._add_component( parent, foo_name, foo_indices, new_obj, saved_connections,
                        saved_upblk_reads, saved_upblk_writes, saved_upblk_calls,
                        saved_func_reads, saved_func_writes, saved_func_calls)

    top._flush_pending_value_connections()
    top._flush_pending_method_connections()
    if check:
      top.check()

  def replace_component_with_obj( top, foo, new_obj, check=True ):
    top._check_called_at_elaborate_top( "replace_component" )

    parent = foo.get_parent_object()
    foo_name    = foo._dsl._my_name
    foo_indices = foo._dsl._my_indices

    saved_connections, saved_upblk_reads, saved_upblk_writes, saved_upblk_calls, \
      saved_func_reads, saved_func_writes, saved_func_calls = top._delete_component( foo )

    # We actually don't need to merge param tree here because when we call
    # _add_component, the parameters stored in parent will be pushed down
    # to new_obj
    top._add_component( parent, foo_name, foo_indices, new_obj, saved_connections,
                        saved_upblk_reads, saved_upblk_writes, saved_upblk_calls,
                        saved_func_reads, saved_func_writes, saved_func_calls)

    top._flush_pending_value_connections()
    top._flush_pending_method_connections()
    if check:
      top.check()

  def add_value_port( top, parent, name, o ):
    top._check_called_at_elaborate_top( "add_port" )

    assert isinstance( o, (InPort, OutPort) )
    # If we are adding field parent.x, we simply reuse the setattr hook
    assert not hasattr( parent, name ), f"Invalid add_value_port call: {parent} already has field {name}!"

    NamedObject._elaborate_stack = [ parent ]
    NamedObject.__setattr__ = NamedObject.__setattr_for_elaborate__
    setattr( parent, name, o )
    del NamedObject.__setattr__
    del NamedObject._elaborate_stack

    top._dsl.all_signals.add( o )
    top._dsl.all_named_objects.add( o )

  def add_connection( top, o1, o2 ):

    # Currently only support connecting signals

    top._check_called_at_elaborate_top( "add_connection" )
    if isinstance( o2, Connectable ): o1, o2 = o2, o1

    assert isinstance( o1, Connectable ), "Cannot connect two non-connectable objects"

    assert isinstance( o1, Signal ), "Currently only support connecting signals/constants"

    assert o1._dsl.elaborate_top is top

    # Add the connection to the correct component
    host1   = o1.get_host_component()
    parent1 = host1.get_parent_object()

    real_host = host1

    if isinstance( o2, Signal ):
      assert o2._dsl.elaborate_top is top

      host2   = o2.get_host_component()
      parent2 = host2.get_parent_object()

      if isinstance( o1, InPort ):
        if isinstance( o2, InPort ):
          if host1 is parent2:
            host1._connect_signal_signal( o1, o2 )
          elif host2 is parent1:
            host2._connect_signal_signal( o1, o2 )
            real_host = host2
          else:
            assert False

        # OutPort-Wire -> same host or wire is out's parent
        elif isinstance( o2, Wire ):
          assert host1 is host2 or host2 is parent1
          host2._connect_signal_signal( o1, o2 )
          real_host = host2

        # InPort-OutPort: enforce same host. it doesn't make much sense
        # to have outport in inport's parent
        elif isinstance( o2, OutPort ):
          assert host1 is host2
          host1._connect_signal_signal( o1, o2 )

        else:
          assert False

      elif isinstance( o1, OutPort ):
        if isinstance( o2, OutPort ):
          if host1 is parent2:
            host1._connect_signal_signal( o1, o2 )
          elif host2 is parent1:
            host2._connect_signal_signal( o1, o2 )
            real_host = host2
          else:
            assert False

        # OutPort-InPort: enforce same host. it doesn't make much sense
        # to have outport in inport's parent
        elif isinstance( o2, InPort ):
          assert host1 is host2
          host1._connect_signal_signal( o1, o2 )

        # OutPort-Wire -> same host or wire is out's parent
        elif isinstance( o2, Wire ):
          assert host1 is host2 or host2 is parent1
          host2._connect_signal_signal( o1, o2 )
          real_host = host2

        else:
          assert False

      elif isinstance( o1, Wire ):
        # Wire-Wire -> same host
        if isinstance( o2, Wire ):
          assert host1 is host2
          host1._connect_signal_signal( o1, o2 )
        # Wire-InPort/OutPort -> same host or wire is in/out's parent
        else:
          assert host1 is host2 or host1 is parent2
          host1._connect_signal_signal( o1, o2 )

      else:
        assert False

      top._dsl.all_adjacency[o1].add(o2)
      top._dsl.all_adjacency[o2].add(o1)
      top._dsl._has_pending_value_connections = True

  def add_connections( s, *args ):
    try:
      top = s._dsl.elaborate_top
    except AttributeError:
      raise NotElaboratedError()

    if len(args) & 1 != 0:
       raise InvalidConnectionError( "Odd number ({}) of objects provided.".format( len(args) ) )

    for i in range(len(args)>>1):
      try:
        s._connect( args[ i<<1 ], args[ (i<<1)+1 ], internal=False )
      except InvalidConnectionError as e:
        raise InvalidConnectionError( "\n- In connect_pair, when connecting {}-th argument to {}-th argument\n{}\n " \
              .format( (i<<1)+1, (i<<1)+2 , e ) )

    for x, adjs in s._dsl.adjacency.items():
      top._dsl.all_adjacency[x].update( adjs )

  # TODO implement everything below and test them

  # Override
  def add_component_by_name( s, name, obj, provided_connections=[] ):
    raise NotImplementedError()

  # Override
  def delete_component( s, name ):
    raise NotImplementedError()

  # Do we still need disconnect and disconnect_pair?



"""
========================================================================
AstHelper.py
========================================================================
This file collects PyMTL ast visitors.

Author : Shunning Jiang
Date   : Jan 17, 2018
"""

import ast
import sys


class DetectVarNames( ast.NodeVisitor ):

  def __init__( self, upblk, obj ):
    self.upblk = upblk
    self.obj = obj
    self.globals = upblk.__globals__
    self.closure = { *upblk.__code__.co_freevars }

    if sys.version_info < (3,8,10000):
      self._get_full_name = self._get_full_name_up_to_py38
    else:
      self._get_full_name = self._get_full_name_starting_py39

  # Helper function to get the full name containing "s"

  def _get_full_name_up_to_py38( self, input_node ):
    node = input_node

    # Store the name/index linearly, and store the corresponding ast nodes linearly
    obj_name = []
    nodelist = []

    # First strip off all slices -- s.x[1][2].y[i][3][2:3]
    slices = []
    while isinstance( node, ast.Subscript ) and isinstance( node.slice, ast.Slice ):
      lower = node.slice.lower
      upper = node.slice.upper
      # If the slice looks like a[i:i+1] where i is variable, I assume it
      # would access the whole variable

      # Shunning: since the closure/global variables can vary across
      # different instances of the same class, we need to cache the name.

      low = up = None

      if isinstance( lower, ast.Num ):
        low = node.slice.lower.n
      elif isinstance( lower, ast.Name ):
        x = lower.id
        if   x in self.globals: low = (False, x)
        elif x in self.closure: low = (True, x)

      if isinstance( upper, ast.Num ):
        up = node.slice.upper.n
      elif isinstance( upper, ast.Name ):
        x = upper.id
        if   x in self.globals: up = (False, x)
        elif x in self.closure: up = (True, x)

      if low is not None and up is not None:
        slices.append( slice(low, up) )
      # FIXME
      # else:

      nodelist.append( node )
      node = node.value

    # Then do the rests.x[1][2].y[i]
    while True:
      num = []
      while isinstance( node, ast.Subscript ) and isinstance( node.slice, ast.Index ):
        v = node.slice.value
        n = "*"

        if isinstance( v, ast.Attribute ): # s.sel, may be constant
          self.visit( v )
        elif isinstance( v, ast.Num ):
          n = v.n
        elif isinstance( v, ast.Name ):
          x = v.id
          if   x in self.globals: n = (False, x)
          elif x in self.closure: n = (True, x)
        elif isinstance( v, ast.Call ): # int(x)
          for x in v.args:
            self.visit(x)

        num.append(n)

        nodelist.append( node )
        node = node.value

      if isinstance( node, ast.Attribute ):
        obj_name.append( (node.attr, num[::-1]) )
      elif isinstance( node, ast.Name ):
        obj_name.append( (node.id, num[::-1]) )
      elif isinstance( node, ast.Subscript ):
        # this is s.x[1:2][1][2] which doesn't make sense
        raise TypeError( f"Having slice in the middle such as s.x[1][1:2][1][2] "
                         f"doesn't make sense at line {input_node.lineno} of "
                         f"update block {self.upblk.__name__} in class {self.obj.__class__}." )
      elif isinstance( node, ast.Call ): # a.b().c()
        # FIXME?
        return None, None
      else:
        assert isinstance( node, ast.Str ) # filter out line_trace
        return None, None

      nodelist.append( node )

      if not hasattr( node, "value" ):
        break
      node = node.value

    if slices:
      assert len(slices) == 1, "Multiple slices at the end of s.%s in update block %s" % \
        ( ".".join( [ obj_name[i][0] + "".join([f"[{x}]" for x in obj_name[i][1]]) for i in range(len(obj_name)) ] ) \
        +  f"[{x[0]}:{x[1]}]", self.upblk.__name__ )

      obj_name[0][1].append( slices[0] )

    obj_name = obj_name[::-1]
    nodelist = nodelist[::-1]
    return obj_name, nodelist

  def _get_full_name_starting_py39( self, input_node ):
    node = input_node

    # Store the name/index linearly, and store the corresponding ast nodes linearly
    obj_name = []
    nodelist = []

    # First strip off all slices -- s.x[1][2].y[i][3][2:3]
    slices = []
    while isinstance( node, ast.Subscript ) and isinstance( node.slice, ast.Slice ):
      lower = node.slice.lower
      upper = node.slice.upper
      # If the slice looks like a[i:i+1] where i is variable, I assume it
      # would access the whole variable

      # Shunning: since the closure/global variables can vary across
      # different instances of the same class, we need to cache the name.

      low = up = None

      if isinstance( lower, ast.Num ):
        low = node.slice.lower.n
      elif isinstance( lower, ast.Name ):
        x = lower.id
        if   x in self.globals: low = (False, x)
        elif x in self.closure: low = (True, x)

      if isinstance( upper, ast.Num ):
        up = node.slice.upper.n
      elif isinstance( upper, ast.Name ):
        x = upper.id
        if   x in self.globals: up = (False, x)
        elif x in self.closure: up = (True, x)

      if low is not None and up is not None:
        slices.append( slice(low, up) )
      # FIXME
      # else:

      nodelist.append( node )
      node = node.value

    # Then do the rests.x[1][2].y[i]
    while True:
      num = []
      while isinstance( node, ast.Subscript ):
        v = node.slice
        n = "*"

        if isinstance( v, ast.Attribute ): # s.sel, may be constant
          self.visit( v )
        elif isinstance( v, ast.Num ):
          n = v.n
        elif isinstance( v, ast.Name ):
          x = v.id
          if   x in self.globals: n = (False, x)
          elif x in self.closure: n = (True, x)
        elif isinstance( v, ast.Call ): # int(x)
          for x in v.args:
            self.visit(x)
        elif isinstance( v, ast.Slice ): # s.sel, may be constant
          raise TypeError( f"Having slice in the middle such as s.x[1][1:2][1][2] "
                           f"doesn't make sense at line {input_node.lineno} of "
                           f"update block {self.upblk.__name__} in class {self.obj.__class__}." )

        num.append(n)

        nodelist.append( node )
        node = node.value

      if   isinstance( node, ast.Attribute ):
        obj_name.append( (node.attr, num[::-1]) )
      elif isinstance( node, ast.Name ):
        obj_name.append( (node.id, num[::-1]) )
      elif isinstance( node, ast.Call ): # a.b().c()
        # FIXME?
        return None, None
      else:
        assert isinstance( node, ast.Str ) # filter out line_trace
        return None, None

      nodelist.append( node )

      if not hasattr( node, "value" ):
        break
      node = node.value

    if slices:
      assert len(slices) == 1, "Multiple slices at the end of s.%s in update block %s" % \
        ( ".".join( [ obj_name[i][0] + "".join([f"[{x}]" for x in obj_name[i][1]]) for i in range(len(obj_name)) ] ) \
        +  f"[{x[0]}:{x[1]}]", self.upblk.__name__ )

      obj_name[0][1].append( slices[0] )

    obj_name = obj_name[::-1]
    nodelist = nodelist[::-1]
    return obj_name, nodelist

class DetectReadsWritesCalls( DetectVarNames ):

  def enter( self, node, read, write, calls ):

    self.read = read
    self.write = write
    self.calls = calls
    self.current_op = None
    self.visit( node )

  def visit_Assign( self, node ):
    for x in node.targets:
      self.visit( x )
    self.visit( node.value )

  def visit_AugAssign( self, node ):
    self.current_op = node.op
    self.visit( node.target )
    self.current_op = None
    self.visit( node.value  )

  def visit_Attribute( self, node ): # s.a.b
    obj_name, nodelist = self._get_full_name( node )
    if not obj_name:  return

    pair = (obj_name, nodelist, self.current_op)

    if   isinstance( node.ctx, ast.Load ):
      self.read.append( pair )
    elif isinstance( node.ctx, ast.Store ):
      self.write.append( pair )
    else:
      raise TypeError( f"Wrong ast node context {type( node.ctx )}" )

  def visit_Subscript( self, node ): # s.a.b[0:3] or s.a.b[0]
    obj_name, nodelist = self._get_full_name( node )
    if not obj_name:  return

    pair = (obj_name, nodelist, self.current_op)

    if   isinstance( node.ctx, ast.Load ):
      self.read.append( pair )
    elif isinstance( node.ctx, ast.Store ):
      self.write.append( pair )
    else:
      raise TypeError( f"Wrong ast node context {type( node.ctx )}" )

    self.visit( node.slice )

  def visit_Call( self, node ):
    obj_name, nodelist = self._get_full_name( node.func )
    if not obj_name:  return

    self.calls.append( (obj_name, nodelist, None) )

    for x in node.args:
      self.visit( x )

  def visit_For( self, node ):
    self.current_op = 'for'
    self.visit( node.target )
    self.current_op = None

    self.visit( node.iter )
    for stmt in node.body:
      self.visit( stmt )
    for stmt in node.orelse:
      self.visit( stmt )

class DetectMethodCalls( DetectVarNames ):

  def enter( self, node, methods ):
    self.methods = methods
    self.visit( node )

  def visit_Call( self, node ):
    obj_name = self.get_full_name( node.func )
    if not obj_name: return # to check node.func.id

    pair = (obj_name, node)

    self.methods.append( pair )

    for x in node.args:
      self.visit( x )

def extract_reads_writes_calls( hostobj, f, tree, read, write, calls ):

  # Traverse the ast to extract variable writes and reads
  # First check and remove @update and empty arguments
  assert isinstance(tree, ast.Module)
  tree = tree.body[0]
  assert isinstance(tree, ast.FunctionDef)

  visitor = DetectReadsWritesCalls( f, hostobj )
  for stmt in tree.body:
    visitor.enter( stmt, read, write, calls )

def get_method_calls( tree, upblk, methods ):
  DetectMethodCalls( upblk, hostobj ).enter( tree, methods )



#=========================================================================
# VerilogTBGenPass.py
#=========================================================================
# Author : Shunning Jiang
# Date   : Mar 18, 2020

from collections import deque

import py

from pymtl3.dsl import MetadataKey
from pymtl3.extra.pypy import custom_exec
from pymtl3.passes.BasePass import BasePass
from pymtl3.passes.rtlir import RTLIRDataType as rdt
from pymtl3.passes.rtlir import RTLIRType as rt

from ..errors import VerilogImportError
from ..util.utility import get_rtype
from .verilog_tbgen_v_template import template as tb_template


class VerilogTBGenPass( BasePass ):
  """ We only generate TB if it is imported or translation-imported """

  # TBGenPass pass public pass data

  #: tbgen case name
  #:
  #: Type: ``str``; input
  #:
  #: Default value: ""
  case_name = MetadataKey(str)

  vtbgen_hooks = MetadataKey(list)

  def __call__( self, top ):
    if not top._dsl.constructed:
      raise VerilogImportError( top,
        f"please elaborate design {top} before applying the TBGen pass!" )

    assert not top.has_metadata( self.vtbgen_hooks )

    tbgen_hooks = []
    top.set_metadata( self.vtbgen_hooks, tbgen_hooks )

    tbgen_components = []

    def traverse_hierarchy( m ):
      if m.has_metadata( self.case_name ) and hasattr(m, '_ports'):
        tbgen_components.append( (m, m.get_metadata( self.case_name )) )
      else:
        for child in m.get_child_components():
          traverse_hierarchy( child )

    traverse_hierarchy( top )

    for x, case_name in tbgen_components:

      signal_decls = []
      task_assign_strs = []
      task_signal_decls = []
      task_check_strs = []
      dut_signal_decls = []

      py_signal_order = []

      for pname, vname, port, is_ifc in x._ports:
        if vname == "reset" or vname == "clk":
          continue

        # Prepare for generating strings
        if   isinstance( port, rt.Port ):  direction = port.get_direction()
        elif isinstance( port, rt.Array ): direction = port.get_sub_type().get_direction()
        else:                              raise Exception( f"unrecognized direction {d}!" )

        p_n_dim, p_rtype = get_rtype( port )
        dtype = p_rtype.get_dtype()
        if   isinstance( dtype, rdt.Vector ): nbits = dtype.get_length()
        elif isinstance( dtype, rdt.Struct ): nbits = dtype.get_class().nbits
        else:                                 raise Exception( f"unrecognized data type {d}!" )

        # Expand pname if we have an array (or arrays) of unexpanded vector ports
        expanded_pname = []
        if isinstance( port, rt.Array ) and len(pname) == 1:
          Q = deque( [ (pname[0], p_n_dim) ] )
          while Q:
            name, dim = Q.popleft()
            if not dim:
              expanded_pname.append(name)
            else:
              for i in range(dim[0]):
                Q.append((f"{name}[{i}]", dim[1:]))
          pname = expanded_pname

        # signal_decls
        signal_decl_indices = " ".join( [ f"[0:{d-1}]" for d in p_n_dim ] )
        signal_decls.append( f"logic [{nbits-1}:0] {vname} {signal_decl_indices}" )

        # dut_signal_decls

        if p_n_dim:
          # https://sutherland-hdl.com/papers/2013-SNUG-SV_Synthesizable-SystemVerilog_paper.pdf
          # chapter 5.2.3
          dut_signal_decls.append( f".{vname}({{ >> {{ {vname} }} }})" )
        else:
          dut_signal_decls.append( f".{vname}({vname})" )

        Q = deque( [ (vname, vname, p_n_dim) ] )
        tot = 0 # This is to keep the same order as the expanded pname list
        while Q:
          name, mangled_name, indices = Q.popleft()
          if not indices:
            pyname = pname[tot]
            if direction == "input":
              task_signal_decls.append( f"input logic [{nbits-1}:0] inp_{mangled_name}" )
              task_assign_strs.append( f"{name} = inp_{mangled_name}")
            else: # output
              task_signal_decls.append( f"input logic [{nbits-1}:0] ref_{mangled_name}" )
              task_check_strs.append( f"`CHECK(lineno, {name}, ref_{mangled_name}, \"{pyname} ({name} in Verilog)\")")
            tot += 1
            py_signal_order.append(pyname)
          else:
            for i in range( indices[0] ):
              Q.append( (f"{name}[{i}]", f"{mangled_name}__{i}", indices[1:]) )

      dut_name = x._ip_cfg.translated_top_module

      with open( f"{dut_name}_{case_name}_tb.v", 'w' ) as output:
        output.write( tb_template.format(
          args_strs         = ",".join([f"a{i}" for i in range(len(task_signal_decls))]),
          harness_name      = dut_name + "_tb",
          signal_decls      = ";\n  ".join(signal_decls), # logic [31:0] xxx [0:3]; -- unpacked array
          task_signal_decls = ",\n    ".join(task_signal_decls), # input logic [31:0] in__x;input logic [31:0] ref_y; -- unpacked ports
          task_assign_strs  = ";\n    ".join(task_assign_strs), # x = in__x; -- unpacked
          task_check_strs   = ";\n    ".join(task_check_strs), # ERR( lineno, 'x', x, ref_x ) -- unpacked
          dut_name          = dut_name,
          dut_clk_decl      = '.clk(clk)' if x._ph_cfg.has_clk else '',
          dut_reset_decl    = '.reset(reset)' if x._ph_cfg.has_reset else '',
          dut_signal_decls  = ",\n    ".join(dut_signal_decls), # logic [31:0] xxx, -- packed array, # .x(x), -- packed array
          cases_file_name   = f"{dut_name}_{case_name}_tb.v.cases",
        ))

      case_file = open( f"{dut_name}_{case_name}_tb.v.cases", "w" )
      tbgen_hooks.append( self.gen_hook_func( top, x, py_signal_order, case_file ) )

  @staticmethod
  def gen_hook_func( top, x, ports, case_file ):
    port_srcs = [ f"'h{{str(x.{p}.to_bits())}}" for p in ports ]

    src =  """
def dump_case():
  if top.sim_cycle_count() > 2: # skip the 2 cycles of reset
    print(f"`T({});", file=case_file, flush=True)
""".format( ",".join(port_srcs) )
    _locals = {}
    custom_exec( py.code.Source(src).compile(), {'top': top, 'x': x, 'case_file': case_file}, _locals)
    return _locals['dump_case']



#=========================================================================
# BehavioralRTLIRTypeCheckL5Pass.py
#=========================================================================
# Author : Peitian Pan
# Date   : March 30, 2019
"""Provide L5 behavioral RTLIR type check pass."""

from pymtl3.passes.rtlir.errors import PyMTLTypeError
from pymtl3.passes.rtlir.rtype import RTLIRDataType as rdt
from pymtl3.passes.rtlir.rtype import RTLIRType as rt

from .BehavioralRTLIRTypeCheckL4Pass import (
    BehavioralRTLIRTypeCheckL4Pass,
    BehavioralRTLIRTypeCheckVisitorL4,
    BehavioralRTLIRTypeEnforcerL4,
)


class BehavioralRTLIRTypeCheckL5Pass( BehavioralRTLIRTypeCheckL4Pass ):
  def get_visitor_class( s ):
    return BehavioralRTLIRTypeCheckVisitorL5

class BehavioralRTLIRTypeCheckVisitorL5( BehavioralRTLIRTypeCheckVisitorL4 ):

  def get_enforce_visitor( s ):
    return BehavioralRTLIRTypeEnforcerL5

  def visit_Attribute( s, node ):
    """Type check an attribute.

    Since only the ports of a subcomponent can be accessed, no explicit
    cross-hierarchy access detection is needed.
    """
    # Attributes of subcomponent can only access ports
    if isinstance( node.value.Type, rt.Component ) and \
       node.value.Type.get_name() != s.component.__class__.__name__:
      if not node.value.Type.has_property( node.attr ):
        raise PyMTLTypeError( s.blk, node.ast,
          f'rt.Component {node.value.Type.get_name()} does not have attribute {node.attr}!' )
      prop = node.value.Type.get_property( node.attr )
      if not rt._is_of_type( prop, ( rt.Port, rt.InterfaceView ) ):
        raise PyMTLTypeError( s.blk, node.ast,
          f'{node.attr} is not a port of {node.value.Type.get_name()} subcomponent!' )
      node.Type = prop
      if isinstance(node.Type, rt.Const) and isinstance(node.Type.get_dtype(), rdt.Vector):
        node._is_explicit = node.Type.get_dtype().is_explicit()
      else:
        node._is_explicit = True
    else:
      super().visit_Attribute( node )

  def visit_Index( s, node ):
    """Type check the index node."""
    if isinstance( node.value.Type, rt.Array ) and \
       isinstance( node.value.Type.get_sub_type(), rt.Component ):
      node.Type = node.value.Type.get_next_dim_type()
      node._is_explicit = True
      s._handle_index_extension( node, node.value, node.idx, 'index' )

    else:
      super().visit_Index( node )

#-------------------------------------------------------------------------
# Enforce types for all terms whose types are inferred (implicit)
#-------------------------------------------------------------------------

class BehavioralRTLIRTypeEnforcerL5( BehavioralRTLIRTypeEnforcerL4 ):
  pass



"""
==========================================================================
 master_minion_ifc.py
==========================================================================
Master/minion send/recv interface implementations at CL and RTL.
 Author: Shunning Jiang
   Date: Jan 28, 2020
"""
from pymtl3 import *

from .ifcs import RecvIfcRTL, SendIfcRTL
from pymtl3.extra import clone_deepcopy

#-------------------------------------------------------------------------
# CL interfaces
#-------------------------------------------------------------------------

# There is no custom connect method in CL ifcs.
# For MasterCL-MinionRTL and MasterRTL-MinionCL connections, we just need
# to connect by name and leverage the custom connect method of the nested
# Send/RecvIfc.

class MasterIfcCL( Interface ):
  def construct( s, ReqType, RespType, *, req=None, req_rdy=None, resp=None, resp_rdy=None ):
    s.ReqType  = ReqType
    s.RespType = RespType
    s.req  = CalleeIfcCL( Type=ReqType,  method=req, rdy=req_rdy )
    s.resp = CalleeIfcCL( Type=RespType, method=resp, rdy=resp_rdy )
  def __str__( s ):
    return f"{s.req}|{s.resp}"

class MinionIfcCL( Interface ):
  def construct( s, ReqType, RespType ):
    s.ReqType  = ReqType
    s.RespType = RespType
    s.req  = CallerIfcCL( Type=ReqType )
    s.resp = CallerIfcCL( Type=RespType )
  def __str__( s ):
    return f"{s.req}|{s.resp}"

#-------------------------------------------------------------------------
# RTL interfaces
#-------------------------------------------------------------------------
# There is no custom connect method in CL ifcs.
# For MasterCL-MinionRTL and MasterRTL-MinionCL connections, we just need
# to connect by name and leverage the custom connect method of the nested
# Send/RecvIfc.

class MasterIfcRTL( Interface ):
  def construct( s, ReqType, RespType ):
    s.ReqType  = ReqType
    s.RespType = RespType
    s.req  = RecvIfcRTL( Type=ReqType )
    s.resp = SendIfcRTL( Type=RespType )
  def __str__( s ):
    return f"{s.req}|{s.resp}"

  def connect( s, other, parent ):
    if isinstance( other, MinionIfcCL ):
      m = ValRdyMasterMinionRTL2CLAdapter( other.ReqType, other.RespType )

      if hasattr( parent, "ValRdyMasterMinionRTL2CLAdapter_count" ):
        count = parent.XcelIfcFL2RTL_count
        setattr( parent, "ValRdyMasterMinionRTL2CLAdapter_count" + str( count ), m )
      else:
        parent.ValRdyMasterMinionRTL2CLAdapter_count = 0
        parent.ValRdyMasterMinionRTL2CLAdapter_0 = m

      s       //= m.left
      m.right //= other
      parent.ValRdyMasterMinionRTL2CLAdapter_count += 1
      return True

    return False

class MinionIfcRTL( Interface ):
  def construct( s, ReqType, RespType ):
    s.ReqType  = ReqType
    s.RespType = RespType
    s.req  = RecvIfcRTL( Type=ReqType )
    s.resp = SendIfcRTL( Type=RespType )
  def __str__( s ):
    return f"{s.req}|{s.resp}"

class ValRdyMasterMinionRTL2CLAdapter( Component ):

  def req_rdy( s ):
    return s.req_entry is not None

  def req( s ):
    assert s.req_entry is not None
    ret = s.req_entry
    s.req_entry = None
    return ret

  def resp_rdy( s ):
    return s.resp_entry is None

  def resp( s, msg ):
    s.resp_entry = clone_deepcopy( msg )

  def construct( s, ReqType, RespType ):
    s.left  = MinionIfcRTL( ReqType, RespType )
    s.right = MasterIfcCL( ReqType, RespType, req=s.req, req_rdy=s.req_rdy, resp=s.resp, resp_rdy=s.resp_rdy )

    # Req side

    s.req_entry = None

    @update_ff
    def up_left_req_rdy():
      s.left.req.rdy <<= (s.req_entry is None)

    @update_once
    def up_left_req_msg():
      if s.req_entry is None:
        if s.left.req.val:
          s.req_entry = clone_deepcopy( s.left.req.msg )

    s.add_constraints(
      U( up_left_req_msg ) < M( s.req ),
      U( up_left_req_msg ) < M( s.req_rdy ),
    )

    # Resp side

    s.resp_entry = None
    s.resp_sent  = Wire()

    @update_once
    def up_right_resp():
      if s.resp_entry is None:
        s.left.resp.val @= 0
      else:
        s.left.resp.val @= 1
        s.left.resp.msg @= s.resp_entry

    @update_ff
    def up_resp_sent():
      s.resp_sent <<= s.left.resp.val & s.left.resp.rdy

    @update_once
    def up_clear():
      if s.resp_sent: # constraints reverse this
        s.resp_entry = None

    s.add_constraints(
      U( up_clear )   < M( s.resp ),
      U( up_clear )   < M( s.resp_rdy ),
      M( s.resp )     < U( up_right_resp ),
      M( s.resp_rdy ) < U( up_right_resp )
    )



from .bits_import import *
from .bits_import import _bitwidths
from .bitstructs import bitstruct, is_bitstruct_class, is_bitstruct_inst, mk_bitstruct
from .helpers import clog2, concat, reduce_and, reduce_or, reduce_xor, sext, trunc, zext



#=========================================================================
# StructuralRTLIRGenL0Pass.py
#=========================================================================

from pymtl3 import MetadataKey
from pymtl3.passes.rtlir.RTLIRPass import RTLIRPass


class StructuralRTLIRGenL0Pass( RTLIRPass ):

  # Pass metadata

  #: RTLIR type of the component
  #:
  #: Type: ``RTLIRType.Component``; output
  rtlir_type = MetadataKey()

  #: RTLIR of all constants that belong to the component
  #:
  #: Type: ``list``; output
  consts = MetadataKey()

  #: RTLIR of all connections in component
  #:
  #: Type: ``list``; output
  connections = MetadataKey()



"""
========================================================================
UnrollTickPass.py
========================================================================
Generate an unrolled tick function

Author : Shunning Jiang
Date   : Dec 26, 2018
"""

import py

from pymtl3.dsl.Connectable import MethodPort
from pymtl3.passes.BasePass import BasePass
from pymtl3.passes.errors import PassOrderError

from ..sim.PrepareSimPass import PrepareSimPass


class UnrollSimPass( PrepareSimPass ):

  @staticmethod
  def gen_tick_function( funclist ):

    # Berkin IlBeyi's recipe ( updated using f-strings and enumerate )
    strs = [ f"_{idx}_{x.__name__}()" for idx, x in enumerate( funclist ) ]

    gen_tick_src = """
def compile_unroll( schedule ):
  {}
  def unrolled():
    pass
    # The code below does the actual calling of update blocks.
    {}
  return unrolled
        """.format( ";".join( [ f"_{idx}_{x.__name__}=schedule[{idx}]"
                                for idx, x in enumerate( funclist ) ] ),
                       "\n    ".join( strs ) )

    l = {}
    exec(py.code.Source( gen_tick_src ).compile(), l)
    return l['compile_unroll']( funclist )

  # Override
  def create_sim_eval_comb( self, top ):
    # FIXME update_once? currently check if the design has method_port
    method_ports = top.get_all_object_filter( lambda x: isinstance( x, MethodPort ) )

    if len(method_ports) == 0: # Pure RTL design, add eval_combinational
      sim_eval_combinational = self.gen_tick_function( top._sched.update_schedule )
    else:
      def sim_eval_combinational():
        raise NotImplementedError(f"top is not a pure RTL design. {'top'+repr(list(method_ports)[0])[1:]} is a method port.")

    top.sim_eval_combinational = sim_eval_combinational

  # Override
  def create_sim_tick( self, top ):
    final_schedule = []
    if not top.get_all_object_filter( lambda x: isinstance( x, MethodPort ) ):
      # Pure RTL -- tick update blocks first
      final_schedule = top._sched.update_schedule[::]

    if self.print_line_trace and hasattr( top, 'line_trace' ):
      final_schedule.append( top.print_line_trace )
    final_schedule += self.collect_ff_funcs( top )
    final_schedule += top._sched.update_schedule
    final_schedule.append( top._sim.check_top_level_inports )
    top.sim_tick = self.gen_tick_function( final_schedule )



#=========================================================================
# BehavioralRTLIRPass_test.py
#=========================================================================
# Author : Peitian Pan
# Date   : Feb 2, 2019
"""Test the behavioral RTLIR passes.

The generation, type check, and visualization passes are invoked. The
results of generation pass are verifed against a reference AST.
"""

# from pymtl3.datatypes import *
from pymtl3.passes.rtlir.behavioral import (
    BehavioralRTLIRGenPass,
    BehavioralRTLIRTypeCheckPass,
    BehavioralRTLIRVisualizationPass,
)
from pymtl3.passes.rtlir.behavioral.BehavioralRTLIR import *
from pymtl3.passes.rtlir.behavioral.test.BehavioralRTLIRL1Pass_test import XFAIL_ON_PY3
from pymtl3.passes.rtlir.errors import PyMTLTypeError
from pymtl3.passes.rtlir.util.test_utility import do_test, expected_failure
from pymtl3.passes.testcases import (
    CaseBits8Bits16MismatchAssignComp,
    CaseBits16ConstAddComp,
    CaseBits16IndexBasicComp,
    CaseBits32Bits64SlicingBasicComp,
    CaseConstSizeSlicingComp,
    CaseFlipFlopAdder,
    CaseForBasicComp,
    CaseForFreeVarStepComp,
    CaseIfBasicComp,
    CaseReducesInx3OutComp,
    CaseSlicingOverIndexComp,
    CaseSubCompAddComp,
    CaseTwoUpblksSliceComp,
)


def local_do_test( m ):
  """Check if generated behavioral RTLIR is the same as reference."""

  ref = m._rtlir_test_ref
  m.elaborate()
  m.apply( BehavioralRTLIRGenPass( m ) )
  m.apply( BehavioralRTLIRTypeCheckPass( m ) )
  m.apply( BehavioralRTLIRVisualizationPass() )

  rtlir_upblks = m.get_metadata( BehavioralRTLIRGenPass.rtlir_upblks )

  for blk in m.get_update_blocks():
    assert rtlir_upblks[ blk ] == ref[ blk.__name__ ]

def test_reduce( do_test ):
  a = CaseReducesInx3OutComp.DUT()
  in_1 = Attribute( Base( a ), 'in_1' )
  in_2 = Attribute( Base( a ), 'in_2' )
  in_3 = Attribute( Base( a ), 'in_3' )
  out = Attribute( Base( a ), 'out' )

  a._rtlir_test_ref = { 'v_reduce' : CombUpblk( 'v_reduce', [
    Assign( [out], BinOp(
      BinOp( Reduce( BitAnd(), in_1 ), BitAnd(), Reduce( BitOr(), in_2 ) ),
      BitOr(), Reduce( BitXor(), in_3 ),
    ), True )
  ] ) }

  do_test( a )

def test_index_basic( do_test ):
  a = CaseBits16IndexBasicComp.DUT()

  a._rtlir_test_ref = { 'index_basic' : CombUpblk( 'index_basic', [
    Assign( [Index( Attribute( Base( a ), 'out' ), Number( 0 ) )],
      BinOp( Index( Attribute( Base( a ), 'in_' ), Number( 0 ) ), Add(),
             Index( Attribute( Base( a ), 'in_' ), Number( 1 ) ) ), True ),
    Assign( [Index( Attribute( Base( a ), 'out' ), Number( 1 ) )],
      BinOp( Index( Attribute( Base( a ), 'in_' ), Number( 2 ) ), Add(),
             Index( Attribute( Base( a ), 'in_' ), Number( 3 ) ) ), True )
  ] ) }

  do_test( a )

def test_mismatch_width_assign( do_test ):
  a = CaseBits8Bits16MismatchAssignComp.DUT()

  a._rtlir_test_ref = { 'mismatch_width_assign' : CombUpblk(
    'mismatch_width_assign', [ Assign(
      [Attribute( Base( a ), 'out' )], Attribute( Base( a ), 'in_' ), True
    )
  ] ) }

  with expected_failure( PyMTLTypeError, "LHS and RHS of assignment" ):
    do_test( a )

def test_slicing_basic( do_test ):
  a = CaseBits32Bits64SlicingBasicComp.DUT()

  a._rtlir_test_ref = { 'slicing_basic' : CombUpblk( 'slicing_basic', [
    Assign( [Slice( Attribute( Base( a ), 'out' ), Number( 0 ), Number( 16 ) )],
      Slice( Attribute( Base( a ), 'in_' ), Number( 16 ), Number( 32 ) ), True ),
    Assign( [Slice( Attribute( Base( a ), 'out' ), Number( 16 ), Number( 32 ) )],
      Slice( Attribute( Base( a ), 'in_' ), Number( 0 ), Number( 16 ) ), True )
  ] ) }

  do_test( a )

def test_bits_basic( do_test ):
  a = CaseBits16ConstAddComp.DUT()

  a._rtlir_test_ref = { 'bits_basic' : CombUpblk( 'bits_basic', [
    Assign( [Attribute( Base( a ), 'out' )],
      BinOp( Attribute( Base( a ), 'in_' ), Add(), SizeCast( 16, Number( 10 ) ) ), True )
  ] ) }

  do_test( a )

def test_index_bits_slicing( do_test ):
  a = CaseSlicingOverIndexComp.DUT()

  a._rtlir_test_ref = { 'index_bits_slicing' : CombUpblk( 'index_bits_slicing', [
    Assign( [Slice(
      Index( Attribute( Base( a ), 'out' ), Number( 0 ) ),
      Number( 0 ), Number( 8 )
      )],
      BinOp(
        BinOp(
          Slice( Index( Attribute( Base( a ), 'in_' ), Number( 1 ) ), Number( 8 ), Number( 16 ) ),
          Add(),
          Slice( Index( Attribute( Base( a ), 'in_' ), Number( 2 ) ), Number( 0 ), Number( 8 ) ),
        ),
        Add(),
        Number( 10 )
      ),
      True
    ),
    Assign(
      [Index( Attribute( Base( a ), 'out' ), Number( 1 ) )],
      BinOp(
        BinOp(
          Slice( Index( Attribute( Base( a ), 'in_' ), Number( 3 ) ), Number( 0 ), Number( 16 ) ),
          Add(),
          Index( Attribute( Base( a ), 'in_' ), Number( 4 ) )
        ),
        Add(),
        Number( 1 )
      ),
      True
    ),
  ] ) }

  do_test( a )

def test_multi_components( do_test ):
  a = CaseSubCompAddComp.DUT()

  a._rtlir_test_ref = { 'multi_components_A' : CombUpblk( 'multi_components_A', [
    Assign( [Attribute( Base( a ), 'out' )],
      BinOp(
        Attribute( Base( a ), 'in_' ),
        Add(),
        Attribute( Attribute( Base( a ), 'b' ), 'out' )
      ),
      True
    )
  ] ) }

  do_test( a )

def test_if_basic( do_test ):
  a = CaseIfBasicComp.DUT()

  a._rtlir_test_ref = {
    'if_basic' : CombUpblk( 'if_basic', [ If(
      Compare( Slice( Attribute( Base( a ), 'in_' ), Number( 0 ), Number( 8 ) ), Eq(), Number( 255 ) ),
      [ Assign( [Attribute( Base( a ), 'out' )], Slice( Attribute( Base( a ), 'in_' ), Number( 8 ), Number( 16 ) ), True ) ],
      [ Assign( [Attribute( Base( a ), 'out' )], Number( 0 ), True ) ]
    )
  ] ) }

  do_test( a )

@XFAIL_ON_PY3
def test_for_basic( do_test ):
  a = CaseForBasicComp.DUT()

  twice_i = BinOp( Number( 2 ), Mult(), LoopVar( 'i' ) )

  a._rtlir_test_ref = {
    'for_basic' : CombUpblk( 'for_basic', [ For(
      LoopVarDecl( 'i' ), Number( 0 ), Number( 8 ), Number( 1 ),
      [ Assign(
          [Slice( Attribute( Base( a ), 'out' ), twice_i, BinOp( twice_i, Add(), Number( 1 ) ) )],
          BinOp(
            Slice( Attribute( Base( a ), 'in_' ), twice_i, BinOp( twice_i, Add(), Number( 1 ) ) ),
            Add(),
            Slice( Attribute( Base( a ), 'in_' ),
              BinOp( twice_i, Add(), Number( 1 ) ),
              BinOp( BinOp( twice_i, Add(), Number( 1 ) ), Add(), Number( 1 ) )
            )
          ),
          True
        )
      ]
    ) ] )
  }

  do_test( a )

def test_for_freevar_step( do_test ):
  a = CaseForFreeVarStepComp.DUT()

  a._rtlir_test_ref = {
    'upblk' : CombUpblk( 'upblk', [ For(
      LoopVarDecl( 'i' ), Number( 0 ), Number( 2 ), FreeVar( 'freevar_at_upblk', 1 ),
      [ Assign(
          [Attribute( Base( a ), 'out' )],
          Slice( Attribute( Base( a ), 'in_' ), Number(0), Number(8) ),
          True
        )
      ]
    ) ] )
  }

  do_test( a )

def test_multi_upblks( do_test ):
  a = CaseTwoUpblksSliceComp.DUT()

  a._rtlir_test_ref = { 'multi_upblks_1' : CombUpblk( 'multi_upblks_1', [
      Assign( [Slice( Attribute( Base( a ), 'out' ), Number(0), Number(4) )], Attribute( Base( a ), 'in_' ), True ),
    ] ),
    'multi_upblks_2' : CombUpblk( 'multi_upblks_2', [
      Assign( [Slice( Attribute( Base( a ), 'out' ), Number(4), Number(8) )], Attribute( Base( a ), 'in_' ), True ),
    ] )
  }

  do_test( a )

def test_ff_upblk( do_test ):
  a = CaseFlipFlopAdder.DUT()
  a._rtlir_test_ref = {
      'update_ff_upblk' : SeqUpblk( 'update_ff_upblk', [
        Assign( [Attribute( Base( a ), 'out0' )], BinOp( Attribute( Base( a ), 'in0' ), Add(), Attribute( Base( a ), 'in1' ) ), False ),
        ] )
  }

  do_test( a )

def test_fixed_size_slice( do_test ):
  a = CaseConstSizeSlicingComp.DUT()
  a._rtlir_test_ref = {
      'upblk' : CombUpblk( 'upblk', [
        For( LoopVarDecl('i'), Number(0), Number(2), Number(1), [
          Assign(
            [Index(Attribute(Base(a), 'out'), LoopVar('i'))],
            Slice(Attribute(Base(a), 'in_'),
                  BinOp(LoopVar('i'), Mult(), Number(8)),
                  BinOp(BinOp(LoopVar('i'), Mult(), Number(8)), Add(), Number(8)),
                  BinOp(LoopVar('i'), Mult(), Number(8)),
                  8,
            ), True)]
        )
      ])
  }

  do_test( a )



from ..BasePass import BasePass
from ..sim.GenDAGPass import GenDAGPass
from ..sim.PrepareSimPass import PrepareSimPass
from ..sim.SimpleSchedulePass import SimpleSchedulePass
from ..sim.WrapGreenletPass import WrapGreenletPass
from ..tracing.CLLineTracePass import CLLineTracePass
from ..tracing.LineTraceParamPass import LineTraceParamPass
from .HeuristicTopoPass import HeuristicTopoPass
from .Mamba2020Pass import Mamba2020Pass
from .UnrollSimPass import UnrollSimPass


class UnrollSim( BasePass ):
  def __init__( s, *, waveform=None, print_line_trace=True, reset_active_high=True ):
    s.waveform = waveform
    s.print_line_trace = print_line_trace
    s.reset_active_high = reset_active_high

  def __call__( s, top ):
    top.elaborate()
    GenDAGPass()( top )
    WrapGreenletPass()( top )
    SimpleSchedulePass()( top )
    UnrollSimPass(print_line_trace=s.print_line_trace,
                  reset_active_high=s.reset_active_high)( top )

class HeuTopoUnrollSim( BasePass ):
  def __init__( s, *, waveform=None, print_line_trace=True, reset_active_high=True ):
    s.waveform = waveform
    s.print_line_trace = print_line_trace
    s.reset_active_high = reset_active_high

  def __call__( s, top ):
    top.elaborate()
    GenDAGPass()( top )
    WrapGreenletPass()( top )
    HeuristicTopoPass(print_line_trace=s.print_line_trace,
                      reset_active_high=s.reset_active_high)( top )

class Mamba2020( BasePass ):
  def __init__( s, *, waveform=None, print_line_trace=True, reset_active_high=True ):
    s.waveform = waveform
    s.print_line_trace = print_line_trace
    s.reset_active_high = reset_active_high

  def __call__( s, top ):
    top.elaborate()
    GenDAGPass()( top )
    WrapGreenletPass()( top )
    if s.print_line_trace:
      CLLineTracePass()( top )
      LineTraceParamPass()( top )
    Mamba2020Pass(print_line_trace=s.print_line_trace,
                  reset_active_high=s.reset_active_high)( top )



#=========================================================================
# VNameMangle_test.py
#=========================================================================
# Author : Peitian Pan
# Date   : June 15, 2019
"""Test the SystemVerilog name mangling."""

from pymtl3.passes.backends.verilog.import_.test.VNameMangle_test import (
    test_interface,
    test_interface_array,
    test_nested_interface,
    test_nested_interface_port_array,
    test_nested_struct,
    test_packed_array_port_array,
    test_port_2d_array,
    test_port_array,
    test_port_single,
    test_struct_port_array,
    test_struct_port_single,
)
from pymtl3.passes.backends.yosys.util.utility import gen_mapped_ports
from pymtl3.passes.rtlir.util.test_utility import do_test


def local_do_test( m ):
  m.elaborate()
  result = gen_mapped_ports( m, {} )
  assert result == m._ref_ports_yosys



#=========================================================================
# VStructuralTranslatorL3.py
#=========================================================================
"""Provide SystemVerilog structural translator implementation."""

from pymtl3.passes.backends.generic.structural.StructuralTranslatorL3 import (
    StructuralTranslatorL3,
)
from pymtl3.passes.rtlir import RTLIRDataType as rdt
from pymtl3.passes.rtlir import RTLIRType as rt

from ...util.utility import make_indent, pretty_concat
from .VStructuralTranslatorL2 import VStructuralTranslatorL2


class VStructuralTranslatorL3(
    VStructuralTranslatorL2, StructuralTranslatorL3 ):

  #-----------------------------------------------------------------------
  # Declarations
  #-----------------------------------------------------------------------

  def rtlir_tr_interface_port_decls( s, port_decls ):
    return sum( port_decls, [] )

  def rtlir_tr_interface_port_decl( s, m, port_id, port_rtype, port_array_type ):
    if isinstance( port_rtype, rt.Port ):
      port_dtype = s.rtlir_data_type_translation( m, port_rtype.get_dtype() )
      if port_array_type is None:
        # Non-array interface port
        return [ {
          'direction' : port_rtype.get_direction(),
          'data_type' : port_dtype['data_type'],
          'packed_type' : port_dtype['packed_type'],
          'unpacked_type' : port_dtype['unpacked_type'],
          'id' : port_id,
        } ]
      else:
        # Array interface port
        return [ {
          'direction' : port_rtype.get_direction(),
          'data_type' : port_dtype['data_type'],
          'packed_type' : port_dtype['packed_type'],
          'unpacked_type' : pretty_concat(port_array_type['unpacked_type'], port_dtype['unpacked_type']),
          'id' : port_id,
        } ]
    else:
      # Nested interface
      dscps = []
      all_properties = port_rtype.get_all_properties_packed()
      for name, _rtype in all_properties:
        if isinstance( _rtype, rt.Array ):
          array_type = _rtype
          rtype = _rtype.get_sub_type()
        else:
          array_type = None
          rtype = _rtype
        # Name-mangle the nested interface
        dscp = s.rtlir_tr_interface_port_decl(m, f'{port_id}__{name}', rtype, array_type)
        # Add unpacked_type of the current interface to the unpacked_type of the port
        for tr in dscp:
          tr['unpacked_type'] = port_array_type['unpacked_type'] + tr['unpacked_type']
        dscps += dscp
      return dscps

  def rtlir_tr_interface_decls( s, ifc_decls ):
    all_decls = sum( ifc_decls, [] )
    make_indent( all_decls, 1 )
    return ',\n'.join( all_decls )

  def rtlir_tr_interface_decl( s, ifc_id, ifc_rtype, array_type, port_decls ):
    decls = []
    for tr in port_decls:
      direc = tr['direction']
      data_type = tr['data_type']
      packed_type = tr['packed_type']
      id = f"{ifc_id}__{tr['id']}"
      unpacked_type = array_type['unpacked_type'] + tr['unpacked_type']
      decls.append(pretty_concat(direc, data_type, packed_type, id, unpacked_type))
    return decls

  #-----------------------------------------------------------------------
  # Signal operations
  #-----------------------------------------------------------------------

  def rtlir_tr_interface_array_index( s, base_signal, index, status ):
    s._rtlir_tr_unpacked_q.append( index )
    return base_signal

  def rtlir_tr_interface_attr( s, base_signal, attr, status ):
    return s._rtlir_tr_process_unpacked(
              f'{base_signal}__{attr}',
              f'{base_signal}__{attr}{{}}',
              status, ('status') )



#=========================================================================
# TranslationImport_closed_loop_component_test.py
#=========================================================================
# Author : Peitian Pan
# Date   : June 6, 2019
"""Closed-loop test cases for translation-import with component."""
import pytest
from hypothesis import HealthCheck, given, reproduce_failure, settings
from hypothesis import strategies as st

from pymtl3.datatypes import Bits1, Bits16, Bits32, bitstruct, clog2, mk_bits
from pymtl3.dsl import Component, InPort, Interface, OutPort, Wire, connect
from pymtl3.passes.testcases import (
    CaseGenericAdderComp,
    CaseGenericConditionalDriveComp,
    CaseGenericMuxComp,
    CaseNestedStructConnectWireComp,
    CaseNestedStructConnectWireSubComp,
    CaseStructConnectWireComp,
)

from ..util.test_utility import closed_loop_component_test

too_slow = [ HealthCheck.too_slow ]

def local_do_test( m, data ):
  closed_loop_component_test( m, data )

# Use @given(st.data()) to draw input vector inside the test function
#  - also note that data should the rightmost argument of the test function
# Set deadline to None to avoid checking how long each test spin is
# Set max_examples to limit the number of attempts after multiple successes
# Suppress `too_slow` healthcheck to avoid marking a long test as failed

@given(st.data())
@settings(deadline = None, max_examples = 5, suppress_health_check = too_slow)
@pytest.mark.parametrize("Type", [Bits16, Bits32])
def test_generic_adder( Type, data ):
  local_do_test( CaseGenericAdderComp.DUT(Type), data )

@given(st.data())
@settings(deadline = None, max_examples = 5, suppress_health_check = too_slow)
@pytest.mark.parametrize(
  "Type, n_ports", [ (Bits16, 2), (Bits16, 4), (Bits32, 2), (Bits32, 4) ]
)
def test_mux( Type, n_ports, data ):
  local_do_test( CaseGenericMuxComp.DUT(Type, n_ports), data )

@given(st.data())
@settings(deadline = None, max_examples = 5, suppress_health_check = too_slow)
def test_struct( data ):
  local_do_test( CaseStructConnectWireComp.DUT(), data )

@given(st.data())
@settings(deadline = None, max_examples = 10, suppress_health_check = too_slow)
def test_nested_struct( data ):
  local_do_test( CaseNestedStructConnectWireComp.DUT(), data )

@given(st.data())
@settings(deadline = None, max_examples = 10, suppress_health_check = too_slow)
def test_subcomp( data ):
  local_do_test( CaseNestedStructConnectWireSubComp.DUT(), data )

# Test contributed by Cheng Tan
@given( st.data() )
@settings( deadline = None, max_examples = 5 )
@pytest.mark.parametrize( "Type", [ Bits16, Bits32 ] )
def test_index_static( Type, data ):
  local_do_test( CaseGenericConditionalDriveComp.DUT(Type), data )
